package de.janschejbal.p2pbluetoothtest.junit.broadcast;

import static org.junit.Assert.assertArrayEquals;
import static org.junit.Assert.assertEquals;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.security.SecureRandom;
import java.util.ArrayList;
import java.util.Random;

import org.junit.Test;

import de.janschejbal.p2pbluetoothtest.broadcast.ChannelState;
import de.janschejbal.p2pbluetoothtest.protocol.base.ProtocolMessage;
import de.janschejbal.p2pbluetoothtest.protocol.base.ProtocolMessageParser;
import de.janschejbal.p2pbluetoothtest.protocol.base.SafeReader;
import de.janschejbal.p2pbluetoothtest.protocol.base.SafeWriter;
import de.janschejbal.p2pbluetoothtest.protocol.broadcast.BroadcastProtocolConnection;
import de.janschejbal.p2pbluetoothtest.protocol.broadcast.EndOfDataMessage;
import de.janschejbal.p2pbluetoothtest.protocol.broadcast.ChannelMessage;
import de.janschejbal.p2pbluetoothtest.protocol.broadcast.DebugMessage;
import de.janschejbal.p2pbluetoothtest.protocol.broadcast.HelloMessage;
import de.janschejbal.p2pbluetoothtest.protocol.broadcast.PayloadMessage;

public class TestMessages {

	private Random rnd = new SecureRandom();
	
	@Test
	public void testChannelState() throws Exception {
		long nodeId = rnd.nextLong();
	
		ChannelState source = new ChannelState(nodeId);
		source.addMessage(11);
		source.addMessage(12);
		source.addMessage(13);
		source.addMessage(16);
		source.addMessage(17);
	
		ChannelState target = new ChannelState(nodeId);
		target.addMessage(12);
		target.addMessage(13);
		target.addMessage(14);
		target.addMessage(15);
		target.addMessage(16);
		target.addMessage(522);
		
		//System.out.println(source.toStringVerbose());
		//System.out.println(target.toStringVerbose());
		assertArrayEquals(new int[] {11, 17}, ChannelState.getMissingRight(source, target).getMessageIDs());
		target.addMessage(523);
		assertArrayEquals(new int[] {17},    ChannelState.getMissingRight(source, target).getMessageIDs());
		String oldString = target.toStringVerbose();
		target.addMessage(11); // should be a noop
		assertEquals(oldString, target.toStringVerbose());
		
		assertArrayEquals(new int[] {14, 15, 522, 523}, ChannelState.getMissingRight(target, source).getMessageIDs());
	}

	@Test
	public void testHelloMessage() throws Exception {
		for (int i = 0; i < 100; i++) {
			long nodeId = rnd.nextLong();
			int minVersion = rnd.nextInt();
			int version = rnd.nextInt();
			int connectionId = rnd.nextInt();
			int remoteInstanceId = rnd.nextInt();
			long remoteTimestamp = rnd.nextLong();
			
			ProtocolMessage sendMsg = new HelloMessage(nodeId, minVersion, version, connectionId, remoteInstanceId, remoteTimestamp);
			HelloMessage parsed = (HelloMessage)simulateXmit(sendMsg, HelloMessage.getParser());
			
			for (int readindex = 0; readindex < 2; readindex ++) {
				assertEquals(nodeId, parsed.getNodeId());
				assertEquals(minVersion, parsed.getMinVersion());
				assertEquals(version, parsed.getVersion());
				assertEquals(connectionId, parsed.getConnectionId());
				assertEquals(remoteInstanceId, parsed.getInstanceId());
				assertEquals(remoteTimestamp, parsed.getTimestamp());
			}
		}
	}

	@Test
	public void testPayloadMessage() throws Exception {
		for (int payloadSize : new int[] {} ) {
			long channelId = rnd.nextLong();
			int messageId = rnd.nextInt(Integer.MAX_VALUE);
			ProtocolMessage sendMsg = new PayloadMessage(channelId, messageId, payloadSize);
			PayloadMessage parsed = (PayloadMessage)simulateXmit(sendMsg, PayloadMessage.getParser());
			assertEquals(payloadSize, parsed.getPayloadSize());
			assertEquals(channelId, parsed.getChannelId());
			assertEquals(messageId, parsed.getMessageId());
		}
	}
	
	@Test
	public void testChannelMessage() throws Exception {
		for (int size = 0; size < 100; size++) {
			ArrayList<ChannelState> originalStates = new ArrayList<ChannelState>();
			for (int i = 0; i < size; i++) {
				ChannelState tmp = new ChannelState(rnd.nextLong());
				for (int j = 0; j < rnd.nextInt(10); j++) tmp.addMessage(rnd.nextInt(700));
				originalStates.add(tmp);
			}
			int tag = rnd.nextBoolean() 
					? BroadcastProtocolConnection.MESSAGE_REQUEST
					: BroadcastProtocolConnection.MESSAGE_STATE;
			
			ChannelMessage parsed = (ChannelMessage)simulateXmit(new ChannelMessage(tag, originalStates),
					ChannelMessage.getParser());
			
			for (int i = 0; i < size; i++) {
				assertEquals(originalStates.get(i).toStringVerbose(), parsed.getChannels().get(i).toStringVerbose());
				//System.out.println(parsed.getChannels().get(i).toStringVerbose());
			}
			//System.out.println();
		}
	}
	
	@Test
	public void testByebyeMessage() throws Exception {
		for (int i = 0; i < 100; i++) {
			int instanceId = rnd.nextInt();
			long timestamp = rnd.nextLong();
			EndOfDataMessage parsed = (EndOfDataMessage)simulateXmit(new EndOfDataMessage(instanceId, timestamp),
					EndOfDataMessage.getParser());
			
			assertEquals(instanceId, parsed.getInstanceId());
			assertEquals(timestamp, parsed.getTimestamp());
		}
	}
	
	@Test
	public void testDebugMessage() throws Exception {
		String ok = "asdfasdfawefacwecfacserhawi4zcq2n43ir qnz34ot q2w4uktr2#+2q~3c5!\"`?=)(/&%$6451 _-<><<|";
		DebugMessage parsed = (DebugMessage)simulateXmit(new DebugMessage(ok), DebugMessage.getParser());
		assertEquals(ok, parsed.getString());
		
		String wrong = "\u007f"+""+""+"\0"+"a";
		parsed = (DebugMessage)simulateXmit(new DebugMessage(wrong), DebugMessage.getParser());
		assertEquals("~~~~a", parsed.getString());
	}
	
	/**
	 * Simulates transmission of a message by sending it to a byte array and parsing it from there. 
	 * @param msg The message to "send"
	 * @return A fresh message parsed from the serialization of msg
	 */
	private ProtocolMessage simulateXmit(ProtocolMessage msg, ProtocolMessageParser parser) throws Exception {
		ByteArrayOutputStream bos = new ByteArrayOutputStream();
		DataOutputStream dos = new DataOutputStream(bos);
		SafeWriter writer = new SafeWriter(dos);
		msg.send(writer);
		assertEquals("Write length mismatch", true, writer.verifyWriteLength());
		dos.flush();
		byte[] bytes = bos.toByteArray();
		dos.close();
		
		ByteArrayInputStream bis = new ByteArrayInputStream(bytes);
		DataInputStream dis = new DataInputStream(bis);
		int tag = dis.readInt();
		int msglen = dis.readInt();
		SafeReader reader = new SafeReader(msglen, dis);
		ProtocolMessage result = parser.parseMessage(tag, msglen, reader);
		dis.close();
		return result;
	}

}
