package de.janschejbal.p2pbluetoothtest.protocol.base;

import java.io.Closeable;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.concurrent.LinkedBlockingQueue;

import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothSocket;
import android.os.Handler;
import android.os.Looper;
import android.os.Message;
import de.janschejbal.p2pbluetoothtest.CONFIG;
import de.janschejbal.p2pbluetoothtest.DataCollectionManager;
import de.janschejbal.p2pbluetoothtest.Logger;
import de.janschejbal.p2pbluetoothtest.util.FileUtils;

public abstract class ProtocolConnection implements Closeable {
	// TODO wakelock?
	
	private final boolean isInitiator;
	protected final BluetoothSocket socket;
	
	private final LinkedBlockingQueue<ProtocolMessage> outputQueue = new LinkedBlockingQueue<ProtocolMessage>();
	private boolean finished = false;
	
	private BluetoothDevice remoteDevice;
	private DataInputStream inputStream;
	private DataOutputStream outputStream;
	private boolean closed;
	private Logger logger;
	private final int connID = (int) (System.currentTimeMillis() % (Integer.MAX_VALUE-1)); // TODO test
	private final String connIDStr = Integer.toHexString(connID);
	private final CloseObserver closeObserver;
	/**
	 * Create a new protocol connection.
	 * The given socket will be closed once this connection finishes, fails, times out or is closed.
	 * YOU MUST CALL {@link #startProcessingThreads()} AT THE *END* OF YOUR CONSTRUCTOR!
	 * 
	 * @param isInitiator true if this is the client-side instance
	 * @param socket the {@link BluetoothSocket} to wrap
	 * @param logger the {@link DataCollectionManager} to use for logging
	 * @param timeout initial timeout in milliseconds
	 * @param observer (optional) a CloseObserver that will be notified when this connection is closed (includes failure, timeout, finished, ...)
	 */
	public ProtocolConnection(boolean isInitiator, BluetoothSocket socket,
			Logger logger, long timeout, CloseObserver observer) {
		this.isInitiator = isInitiator;
		this.socket = socket;
		this.logger = logger;
		this.closeObserver = observer;
		
		log("init", isInitiator ? "out" : "in");
		

		try {
			this.remoteDevice = socket.getRemoteDevice();
			log("remote name", socket.getRemoteDevice().getName());
			inputStream = new DataInputStream(socket.getInputStream());
			outputStream = new DataOutputStream(socket.getOutputStream());
		} catch (IOException e) {
			log("init failed");
			close();
			return;
		}
		
		setTimeout(timeout);
	}
	
	/**
	 * Call this after your instance is initialized.
	 * Attempting to do this in the {@link ProtocolConnection} constructor leads to very interesting effects:
	 * The thread starts and uses uninitialized fields of the subclass, which *then* get initialized.
	 */
	protected void startProcessingThreads() {
		inputProcessingThread.start();
		outputProcessingThread.start();
	}
	
	/**
	 * This method gets called on the connection's input processing thread
	 * each time a message header is received. 
	 * Override it to implement your protocol handler.
	 * It is your job to either read exactly length bytes (i.e. the message) from inputStream and return true
	 * or read 0 bytes and return false.
	 * 
	 * <p>Use {@link #sendMessage(ProtocolMessage)} to send messages,
	 * {@link #done()} to indicate you expect no further messages
	 * (i.e. stop reading and close the connection as soon as the output queue has been processed),
	 * and {@link #close()} to immediately shut down the connection.</p>
	 * 
	 * @param tag the tag of the message - use this to select the correct constructor/factory to use
	 * @param length the length of the message
	 * @param inputStream the input stream from which you may read the body of your message
	 * @return true if the message was read, false if the message should be skipped
	 * @throws IOException 
	 */
	protected abstract boolean handleMessage(int tag, int length, SafeReader reader) throws IOException;
	
	/**
	 * This method will be called once on the connection's input processing thread
	 * immediately when the processing thread is started.
	 * You can use it to send the first protocol message using {@link #sendMessage(ProtocolMessage)}.
	 * You may want to check {@link #isInitiator()} to determine who initiated the connection.
	 */
	protected abstract void initiateConnection();
	
	/**
	 * This method will be called when the connection is "done" (i.e. has received all expected messages).
	 */
	protected void onDone() {};

	/**
	 * This method will be called when the connection is closed.
	 * You may also pass a CloseObserver in the constructor.
	 */
	protected void onClose() {};
	
	/**
	 * Call this method to add a message to the outgoing message queue.
	 * (The output processing thread will retrieve and send it to the output stream.)
	 * @param msg the message to be sent (null messages will be silently ignored)
	 */
	protected void sendMessage(ProtocolMessage msg) {
		if (msg == null) return;
		outputQueue.add(msg);
	}

	/**
	 * Sets the timeout for this connection, restarting the timer.
	 * {@link #close()} is called when the connection times out.
	 * You may want to set a new timeout after a message was received.
	 * @param millis milliseconds (of uptime) from now until the connection times out
	 */
	public void setTimeout(long millis) {
		timeoutHandler.removeMessages(0);
		timeoutHandler.sendEmptyMessageDelayed(0, millis);
	}
	
	/**
	 * Call this when no further messages are expected.
	 * This sets the "finished" flag, meaning the input thread will terminate after the current iteration.
	 * The output thread will send any remaining messages in the output queue and terminate.
	 */
	public void done() {
		log("done");
		finished = true;
		outputQueue.add(QUEUE_INTERRUPT_MARKER);
		onDone();
	}

	/**
	 * Immediately closes the protocol handler.
	 * The shutdown flag is set, the streams and the socket are closed,
	 * and the input/output processing threads are interrupted (but not stopped hard).
	 */
	public void close() {
		synchronized (this) {
			if (isClosed()) return;
			closed = true;
		}
		log("closing");			
		timeoutHandler.removeCallbacksAndMessages(null);
		FileUtils.closeSilent(inputStream);
		FileUtils.closeSilent(outputStream);
		FileUtils.closeSilent(socket);
		outputQueue.clear();
		if (inputProcessingThread != null) inputProcessingThread.interrupt();
		if (outputProcessingThread != null) outputProcessingThread.interrupt();
		onClose();
		if (closeObserver != null) closeObserver.onConnectionClose(this);
		synchronized (this) {
			notifyAll();
		}
	}

	private Handler timeoutHandler = new Handler(Looper.getMainLooper(), new Handler.Callback() {
		@Override
		public boolean handleMessage(Message msg) {
			log("timeout");
			ProtocolConnection.this.close();
			return true;
		}
	});
	
	private final Thread inputProcessingThread = new Thread() {
		public void run() {
			initiateConnection();
			while (!closed && !finished) {
				try {
					int tag = inputStream.readInt();
					int length = inputStream.readInt();
					if (length > CONFIG.MAX_MESSAGE_LENGTH || length < 0) throw new ProtocolException("Invalid length");
					SafeReader safeInput = new SafeReader(length, inputStream);
					if (!ProtocolConnection.this.handleMessage(tag, length, safeInput)) safeInput.skipBytes(length);
					if (!safeInput.verifyReadLength()) throw new ProtocolException("Message not read completely");
				} catch (Exception e) {
					if (!finished) log("input exception", e.toString());
					ProtocolConnection.this.close();
				}
			}
		}
	};
	
	private final Thread outputProcessingThread = new Thread() {
		public void run() {
			while (!closed && (!finished || !outputQueue.isEmpty())) {
				try {
					ProtocolMessage msg = outputQueue.take();
					if (msg == QUEUE_INTERRUPT_MARKER) continue; // provides a way to "interrupt" only the queue-waiting
					SafeWriter safeOutput = new SafeWriter(outputStream);
					msg.send(safeOutput);
					if (!safeOutput.verifyWriteLength()) throw new ProtocolException("Message not written completely");
				} catch (InterruptedException e) {
					// ignore, shutdown = false will stop loop
				} catch (Exception e) {
					if (!finished) log("output exception", e.toString());
					break;
				}
			}
			ProtocolConnection.this.close();
		};
	};
	
	/**
	 * Logs information regarding this connection, adding the connection ID
	 * @param parts a maximum of 5 additional message parts
	 */
	public void log(Object... parts) {
		// Attempting to pass a connID *and* an array of strings/objects will log the Array's toString()
		switch (parts.length) {
		case 0:  logger.log(getTag(), connIDStr); break;
		case 1:  logger.log(getTag(), connIDStr, parts[0]); break;
		case 2:  logger.log(getTag(), connIDStr, parts[0], parts[1]); break;
		case 3:  logger.log(getTag(), connIDStr, parts[0], parts[1], parts[2]); break;
		case 4:  logger.log(getTag(), connIDStr, parts[0], parts[1], parts[2], parts[3]); break;
		case 5:  logger.log(getTag(), connIDStr, parts[0], parts[1], parts[2], parts[3], parts[4]); break;
		default: throw new RuntimeException("ProtocolConnection.log() cannot handle so many parameters");
		}		
	}
	
	/**
	 * @return the debug tag
	 */
	protected abstract String getTag();
	
	public BluetoothDevice getRemoteDevice() {
		return remoteDevice;
	}

	public int getConnID() {
		return connID;
	}

	public boolean isInitiator() {
		return isInitiator;
	}
	
	public boolean isDone() {
		return finished;
	}
	
	public boolean isClosed() {
		return closed;
	}
	
	public interface CloseObserver {
		public void onConnectionClose(ProtocolConnection protocolConnection);
	}

	/**
	 * Dummy protocol message used to interrupt waiting on the queue. (Will be discarded)
	 */
	private final static ProtocolMessage QUEUE_INTERRUPT_MARKER = new ProtocolMessage() {
		@Override public void send(SafeWriter sink) throws IOException {}
	};
}
