package de.janschejbal.p2pbluetoothtest.protocol.base;

import java.io.DataInput;
import java.io.IOException;

public class SafeReader implements DataInput {

	private final int length;
	private final DataInput stream;
	private int processed = 0;
	
	public SafeReader(int length, DataInput stream) {
		if (length < 0) throw new RuntimeException("Negative length");
		this.length = length;
		this.stream = stream;
	}
	
	/**
	 * Reads and discards the rest of the message (if any)
	 * @throws IOException 
	 * @returns true if there was an overhang
	 */
	public boolean chompOverhang() throws IOException {
		int overhang = length - processed;
		if (overhang <= 0) return false;
		skipBytes(overhang);
		return true;
	}
	
	/**
	 * Checks if the read was correct/complete
	 * @return true iff exactly length bytes were read.
	 */
	public boolean verifyReadLength() {
		return processed == length;
	}
	

	private void checkRead(int amount) {
		processed += amount;
		if (processed > length) throw new RuntimeException("Attempted to read beyond message length");
	}

	@Override
	public void readFully(byte[] dst) throws IOException {
		checkRead(dst.length);		
		stream.readFully(dst);
	}

	@Override
	public void readFully(byte[] dst, int offset, int count) throws IOException {
		checkRead(count);		
		stream.readFully(dst, offset, count);
	}

	@Override
	public boolean readBoolean() throws IOException {
		checkRead(1);
		return stream.readBoolean();
	}

	@Override
	public char readChar() throws IOException {
		checkRead(2);
		return stream.readChar();
	}

	@Override
	public byte readByte() throws IOException {
		checkRead(1);
		return stream.readByte();
	}

	@Override
	public int readUnsignedByte() throws IOException {
		checkRead(1);
		return stream.readUnsignedByte();
	}

	@Override
	public short readShort() throws IOException {
		checkRead(2);		
		return stream.readShort();
	}

	@Override
	public int readUnsignedShort() throws IOException {
		checkRead(2);		
		return stream.readUnsignedShort();
	}

	@Override
	public int readInt() throws IOException {
		checkRead(4);		
		return stream.readInt();
	}

	@Override
	public long readLong() throws IOException {
		checkRead(8);		
		return stream.readLong();
	}

	@Override
	public double readDouble() throws IOException {
		checkRead(8);
		return stream.readDouble();
	}

	@Override
	public float readFloat() throws IOException {
		checkRead(4);
		return stream.readFloat();
	}

	@Override
	public int skipBytes(int count) throws IOException {
		checkRead(count);
		int skipped = stream.skipBytes(count);
		// Should not happen per documentation, but might happen in certain corner cases
		if (skipped != count) throw new IOException("skipBytes failed to skip correct number of bytes");
		return skipped;
	}

	@Override
	public String readLine() throws IOException {
		throw new UnsupportedOperationException("Not implemented");
	}

	@Override
	public String readUTF() throws IOException {
		throw new UnsupportedOperationException("Not implemented");
	}

}
