package de.janschejbal.p2pbluetoothtest.protocol.simple;

import java.io.IOException;

import android.bluetooth.BluetoothSocket;
import de.janschejbal.p2pbluetoothtest.Logger;
import de.janschejbal.p2pbluetoothtest.protocol.base.ProtocolConnection;
import de.janschejbal.p2pbluetoothtest.protocol.base.SafeReader;


/**
 * Simple protocol for measurement purposes.
 * 
 * Sends 5 messages with 100 byte (as responses to previous message, initiator sends first)
 *   -- messages 0,1,2,3,4
 * Sends 5 messages with 10 kB (as responses)
 *   -- messages 5,6,7,8,9
 * Sends 3 messages with 100 kB (as responses)
 *   -- messages 10,11,12
 * Sends 1 message with 0 bytes (as responses)
 *   -- message 13
 * Sends 5 messages with 100 kB (parallelized, starting right after handling 0-byte message)
 *
 */
public class SimpleProtocolConnection extends ProtocolConnection {
	
	private static final long SMALL_TIMEOUT = 20*1000;
	private static final long LARGE_TIMEOUT = 60*1000;
	
	
	private int currentMsg = 0;
	
	/** number of messages from the final, asyncronous exchange already received */;
	private int asyncMsgCount = 0;
	
	protected String getTag() { return "SimpleProtocolConnection"; }
	
	public SimpleProtocolConnection(boolean isInitiator, BluetoothSocket socket,
			Logger logger, CloseObserver observer) {
		super(isInitiator, socket, logger, SMALL_TIMEOUT, observer);
		startProcessingThreads();
	}

	@Override
	protected boolean handleMessage(int tag, int length, SafeReader reader) throws IOException {
		reader.readFully(new byte[length]); // Read and discard body
		processMessage(tag);
		return true;
	}
	

	protected void processMessage(int tag) {
		if (tag != 14) {
			sendNext();
		} else {
			asyncMsgCount++;
			if (asyncMsgCount == 5) done();
		}
	}

	@Override
	protected void initiateConnection() {
		if (isInitiator()) sendNext();
	}
	
	private void sendNext() {
		log("sendNext entering", "currentMsg", currentMsg, "thread", Thread.currentThread().getId());
		if (currentMsg < 5) sendRandomMessage(100);
		else if (currentMsg < 10) sendRandomMessage(10*1024);
		else if (currentMsg < 13) sendRandomMessage(100*1024);
		else if (currentMsg == 13) sendMessage(new RandomBodyMessage(13, 0));
		
		if ((currentMsg == 13 && !isInitiator()) || currentMsg == 14) {
			currentMsg = 14;
			sendRandomMessage(100*1024);
			sendRandomMessage(100*1024);
			sendRandomMessage(100*1024);
			sendRandomMessage(100*1024);
			sendRandomMessage(100*1024);
		}
		
		setTimeout(currentMsg < 14 ? SMALL_TIMEOUT : LARGE_TIMEOUT);		
		currentMsg++;
		log("sendNext leaving", "currentMsg", currentMsg, "thread", Thread.currentThread().getId());
	}

	private void sendRandomMessage(int size) {
		sendMessage(new RandomBodyMessage(currentMsg, size));
	}

}
