package de.janschejbal.p2pbluetoothtest.util;

import java.util.HashMap;
import java.util.UUID;

import android.app.AlarmManager;
import android.app.PendingIntent;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.net.Uri;
import android.os.SystemClock;

/**
 * Turns the pain in the ass named AlarmManager into something usable.
 */
public class AlarmHelper {

	private static final String EXTRA_REPEATING = "Repeating";
	private final Context context;
	private final AlarmManager alarmManager;
	private final BroadcastReceiver receiver;
	private boolean closed;
	private long nextId = 0;
	
	private final String actionName;
	
	private HashMap<Long, AlarmRunnable> runnableMap = new HashMap<Long, AlarmRunnable>();
	
	public AlarmHelper(Context context) {
		this.context = context;
		this.alarmManager = (AlarmManager)context.getSystemService(Context.ALARM_SERVICE);
		this.actionName = AlarmHelper.class.getCanonicalName() +"."+ UUID.randomUUID();
		
		receiver = new BroadcastReceiver() {
			@Override
			public void onReceive(Context context, Intent intent) {
				//Log.i("Alarm", "receive " + intent);
				handleBroadcast(intent);
			}
		};
		
		try {
			IntentFilter filter = new IntentFilter(actionName);
			filter.addDataScheme("data");
			filter.addDataType("*/*");
			context.registerReceiver(receiver, filter);
		} catch (Exception e) {
			throw new RuntimeException("AlarmHelperFail", e);
		}
	}
	

	public synchronized long setAlarm(boolean wakeup, long delay, AlarmRunnable runnable) {
		if (closed) return -1;
		//Log.i("Alarm", "set d" + delay + " w" + wakeup + " id"+nextId);
		runnableMap.put(nextId, runnable);
		
		alarmManager.set(wakeup? AlarmManager.ELAPSED_REALTIME_WAKEUP : AlarmManager.ELAPSED_REALTIME,
					SystemClock.elapsedRealtime()+delay,
					getIntent(nextId, false)
				);
		return nextId++;
	}

	public synchronized long setRepeatingAlarm(boolean wakeup, long delay, long repeatInterval,
			AlarmRunnable runnable) {
		if (closed) return -1;
		
		runnableMap.put(nextId, runnable);
		
		alarmManager.setRepeating(wakeup? AlarmManager.ELAPSED_REALTIME_WAKEUP : AlarmManager.ELAPSED_REALTIME,
					SystemClock.elapsedRealtime()+delay,
					repeatInterval,
					getIntent(nextId, true)
				);
		return nextId++;
	}
	
	public synchronized void cancelAlarm(long id) {
		//Log.i("Alarm", "cancel id" + id);

		if (closed) return;

		alarmManager.cancel(getIntent(id, false));
		runnableMap.remove(id);
	}
	
	private PendingIntent getIntent(long id, boolean repeating) {
		// Putting the ID in the data makes sure the intents for the different IDs are not considered equal
		// Otherwise, cancelling individual alarms would be impossible
		Intent intent = new Intent(actionName);
		intent.setDataAndType(Uri.parse("data:," + id), "text/plain");
		if (repeating) intent.putExtra(EXTRA_REPEATING, true);
		return PendingIntent.getBroadcast(context, 0, intent, 0);
	}

	private void handleBroadcast(Intent intent) {
		AlarmRunnable runnable;
		
		synchronized (this) {
			if (closed) return;
			if (!actionName.equals(intent.getAction())) return;
			long id = Long.parseLong(intent.getData().toString().replace("data:,", ""));
			runnable = runnableMap.get(id);
			if (!intent.getBooleanExtra(EXTRA_REPEATING, false)) runnableMap.remove(id);
		}
		
		if (runnable != null) runnable.handleAlarm(intent.getIntExtra(Intent.EXTRA_ALARM_COUNT, 1));
	}

	public synchronized void cancelAll() {
		for (Long id : runnableMap.keySet()) {
			alarmManager.cancel(getIntent(id, false));
		}
		runnableMap.clear();
	}


	public synchronized void close() {
		closed = true;
		context.unregisterReceiver(receiver);
		cancelAll();
		runnableMap = null;
	}

	public interface AlarmRunnable {
		/**
		 * Runs the runnable
		 * @param count alarm repeat count (always 1 for one-shot alarms) - see Intent.EXTRA_ALARM_COUNT
		 */
		public void handleAlarm(int count);
	}

}
