package de.janschejbal.p2pbluetoothtest;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.zip.GZIPOutputStream;


import org.apache.http.entity.mime.MultipartEntity;
import org.apache.http.entity.mime.content.FileBody;

import android.annotation.SuppressLint;
import android.content.Context;
import android.content.Intent;
import android.net.ConnectivityManager;
import android.net.Uri;
import android.net.wifi.WifiManager;
import android.net.wifi.WifiManager.WifiLock;
import android.os.AsyncTask;
import android.os.Build;
import android.os.Handler;
import android.os.Handler.Callback;
import android.os.Message;
import android.os.StrictMode;
import android.os.StrictMode.ThreadPolicy;
import android.os.StrictMode.VmPolicy;
import android.os.SystemClock;
import android.telephony.TelephonyManager;
import android.util.Log;
import de.janschejbal.p2pbluetoothtest.util.AlarmHelper;
import de.janschejbal.p2pbluetoothtest.util.AlarmHelper.AlarmRunnable;
import de.janschejbal.p2pbluetoothtest.util.CryptoHelper;
import de.janschejbal.p2pbluetoothtest.util.FileUtils;
import de.janschejbal.p2pbluetoothtest.util.HttpsClient;
import de.janschejbal.p2pbluetoothtest.util.SimpleReceiver;
import de.janschejbal.p2pbluetoothtest.util.SntpClient;

/**
 * Handles logs and uploads
 */
public class LoggingManager extends SimpleReceiver implements Logger {
	private final BackgroundService service;
	private final static String TAG = "LoggingManager";
	
	private static final String SERVER_MESSAGE_FILENAME = "message";
	private static final String NTP_SERVER = "0.de.pool.ntp.org";
	
	// UDP logging
	private DatagramSocket logsock;
	private final ConnectivityManager connManager;
	private final HttpsClient httpsClient;

	private OutputStreamWriter writer;
	private long logWrittenLines = 0;
	
	private final File baseLogPath;
	private final File workingLogPath;
	private final File uploadLogPath;
	
	private UploadThread uploadThread;
	private boolean uploading = false;
	
	/** Indicates that the logging manager is closing and no new logs may be created */
	private boolean terminate;
	private int uploadCount = 0;
	private long uploadSize = 0;
	private long uploadOldest;
	private Handler uploadStateHandler = new Handler(new Callback() {
		@Override
		public boolean handleMessage(Message msg) {
			service.updateStatus();
			return true;
		}
	});
	
	private String serverMessageUrl = null;
	private String serverMessageText = null;
	
	private final AlarmHelper alarm;
	private long logTimeoutAlarmId = -1;
	private long logUploadAlarmId = -1;
	
	/** 
	 * Used to detect that no reboot occurred between logs.
	 * If two logs have the same ID, it can be assumed that no reboot happened in between
	 * (the reverse is not true: the ID can change without a reboot due to crashes/updates, etc).
	 * Thus, if we have a NTP sync in a log with a certain ID,
	 * we can apply it to all logs with the same ID.
	 */
	private final long uptimeId;
	
	public LoggingManager(BackgroundService service) {
		super(service, Intent.ACTION_SHUTDOWN);
		
		this.service = service;
		this.alarm = new AlarmHelper(service);
		
		uptimeId = CryptoHelper.random.nextLong();
		
		connManager = (ConnectivityManager) service.getSystemService(Context.CONNECTIVITY_SERVICE);
		httpsClient = new HttpsClient(service, false);
		
		baseLogPath = new File(service.getFilesDir(), "logs");
		workingLogPath = new File(baseLogPath, "working");
		uploadLogPath  = new File(baseLogPath, "upload");
		workingLogPath.mkdirs();
		uploadLogPath.mkdirs();
				
		try {
			if (CONFIG.LOGGING_UDP) {
				StrictMode.setVmPolicy(VmPolicy.LAX);
				StrictMode.setThreadPolicy(ThreadPolicy.LAX);
				logsock = new DatagramSocket(11111);
			}
			//openLog();
		} catch (IOException e) {
			Log.e(TAG, "Cannot open log file/socket", e);
		}
		
		String msg = FileUtils.readString(service, SERVER_MESSAGE_FILENAME);
		if (msg != null) {
			String[] parts = msg.split("\t");
			if (parts.length == 2) {
				serverMessageUrl = parts[0];
				serverMessageText = parts[1];
			}
		}
		
		updateFileStats();
		uploadThread = new UploadThread();
		uploadThread.start();
	}
	
	/**
	 * Logs the given data only if a log file is already open
	 * @param tag
	 * @param msgElements
	 */
	public synchronized void logIfOpen(String tag, Object... msgElements) {
		if (writer != null) log(tag, msgElements);
	}
	
	/**
	 * Logs the given data, opening a new log file if needed
	 * @param tag
	 * @param msgElements
	 */
	public synchronized void log(String tag, Object... msgElements) {
		if (writer == null) openLog();
		
		
		StringBuilder logstring = new StringBuilder();
		logstring.append(System.currentTimeMillis());
		logstring.append("\t").append(SystemClock.elapsedRealtime());
		logstring.append("\t").append(tag);
		for (int i = 0; i < msgElements.length; i++) {
			logstring.append("\t").append(escapeLog(msgElements[i]));
		}
		logstring.append("\n");

		if (CONFIG.LOGGING_ADB) Log.v("Log", logstring.toString().replace("\t", "    "));
		
		try {
			writer.write(logstring.toString());
		} catch (Exception e) {
			Log.e(TAG, "Failed to write collection log"  +e);
		}
	
		// UDP logging
		if (CONFIG.LOGGING_UDP) {
			if (hasWifi()) {
				try {
					byte[] data = logstring.toString().getBytes();
					logsock.send(new DatagramPacket(data, data.length, CONFIG.UDP_LOG_TARGET));
				} catch (Exception e) {
					Log.e(TAG, "Failed to UDP-write collection log"  +e);
				}
			}
		}
		
		logWrittenLines++;
		
		if (logWrittenLines >= CONFIG.MAX_LOG_LINES) {
			logWrittenLines = 0; // avoid infinite recursion
			log(TAG, "Log reached max lines");
			closeLog();
		}
	}

	private static String escapeLog(Object msg) {
		if (msg == null) return "[NULL]";
		String str = (msg instanceof String) ? (String)msg : msg.toString();
		return str.replace("\n", "\\n").replace("\t", "\\t");
	}
	
	/**
	 * Logs general info about the device, including the node ID.
	 * Called as soon as a log is created.
	 */
	private synchronized void gatherDeviceInfo() {		
		
		log("APP", "version", service.getVersion());
		log("APP", "builddate", service.getBuildDate());
		log("APP", "prefix", CONFIG.BT_NAME_PREFIX);
		
		log("UptimeID", uptimeId);
		fetchTimeInfo();

		
		log("DEVICE", "free space", baseLogPath.getUsableSpace());
		
		log("DEVICE", "P2P.NodeID", service.getNodeID(), CONFIG.BT_NAME_PREFIX);
		

		TelephonyManager phone = (TelephonyManager)(service.getSystemService(Context.TELEPHONY_SERVICE));
		log("DEVICE", "NetworkType", String.valueOf(phone.getNetworkType()));
		log("DEVICE", "PhoneType", String.valueOf(phone.getPhoneType()));

		// Platform
		log("DEVICE", "Build.BOARD", Build.BOARD);
		log("DEVICE", "Build.BRAND", Build.BRAND);
		log("DEVICE", "Build.DEVICE", Build.DEVICE);
		log("DEVICE", "Build.DISPLAY", Build.DISPLAY);
		log("DEVICE", "Build.HARDWARE", Build.HARDWARE);
		log("DEVICE", "Build.MANUFACTURER", Build.MANUFACTURER);
		log("DEVICE", "Build.MODEL", Build.MODEL);
		log("DEVICE", "Build.PRODUCT", Build.PRODUCT);
		log("DEVICE", "Build.TAGS", Build.TAGS);
		log("DEVICE", "Build.TYPE", Build.TYPE);
		log("DEVICE", "Build.VERSION.CODENAME", Build.VERSION.CODENAME);
		log("DEVICE", "Build.VERSION.INCREMENTAL", Build.VERSION.INCREMENTAL);
		log("DEVICE", "Build.VERSION.RELEASE", Build.VERSION.RELEASE);
		log("DEVICE", "Build.VERSION.SDK", Build.VERSION.SDK);
		log("DEVICE", "Build.VERSION.SDK_INT", Build.VERSION.SDK_INT);
		if (service.isBluetoothBuggy()) {
			log("DEVICE", "BLUETOOTHBUG");			
		}
	}

	private synchronized void fetchTimeInfo() {
		new AsyncTask<Void, Void, Void>() {
			@SuppressLint("InlinedApi")
			private final static int lockmode = WifiManager.WIFI_MODE_FULL_HIGH_PERF;
			
			@Override
			protected Void doInBackground(Void... params) {
				if (!hasWifi()) return null;
				WifiLock lock = null;
				
				try {
					long debugtime = SystemClock.elapsedRealtime();
					Log.d("SNTP wifi lock", "Getting Wifi lock");
					WifiManager wifi = (WifiManager)service.getSystemService(Context.WIFI_SERVICE);
					lock = wifi.createWifiLock(lockmode, "de.janschejbal.p2pbluetoothtest.LoggingManager.SNTPwifilock-"+SystemClock.elapsedRealtime());
					lock.acquire();
					Log.d("SNTP wifi lock", "Got Wifi lock after " + (SystemClock.elapsedRealtime() - debugtime));
				} catch (Exception e) {
					// Officially, WIFI_MODE_FULL_HIGH_PERF is only supported since API 12.
					// Inofficially, most Gingerbread phones support it.
					// Thus, we try to use it and ignore failures.
					Log.d("SNTP wifi lock", "failed", e);
				}
				
				try { Thread.sleep(1000); } catch (InterruptedException e) {} // ensure radio had time to wake up
				
				SntpClient ntp = new SntpClient();
				if (ntp.requestTime(NTP_SERVER, 1000)) {
					logIfOpen("NTPTIME", ntp.getNtpTime(), ntp.getNtpTimeReference(), ntp.getRoundTripTime());					
				} else {
					logIfOpen("NTPFAIL");
				}
				
				try { lock.release(); } catch (Exception e) {}
				
				return null;
			}
		}.execute((Void[])null);
	}

	private synchronized void openLog() {
		if (terminate) {
			Log.e(TAG, "Cannot open new log, LoggingManager is terminating");
			return;
		}
		
		try {
			String timestamp = Long.toString(System.currentTimeMillis());
			FileOutputStream fOut = new FileOutputStream(new File(workingLogPath, timestamp+".log.gz"));
			GZIPOutputStream gzOut = new GZIPOutputStream(new BufferedOutputStream(fOut));
			this.writer = new OutputStreamWriter(gzOut);
			logWrittenLines = 0;
			log(TAG, "Opening new log", timestamp);
			gatherDeviceInfo();
			
			if (logTimeoutAlarmId  != -1) alarm.cancelAlarm(logTimeoutAlarmId);
			logTimeoutAlarmId = alarm.setAlarm(false, CONFIG.MAX_LOG_TIME, new AlarmRunnable() {
				@Override
				public void handleAlarm(int count) {
					Log.i(TAG, "Log time exceeded, closing log");
					closeLog();
				}
			});
		} catch (IOException e) {
			Log.e(TAG, "Cannot open new log", e);
		}
		
	}
	
	/**
	 * Closes the current log file.
	 * A new log file will be opened automatically when logging is required.
	 * @param suppressUpload if true, no upload attempt will be started (e.g. during shutdown)
	 */
	public synchronized void closeLog(boolean suppressUpload) {
		if (writer != null) {
			log(TAG, "Closing log");
			FileUtils.closeSilent(writer);
			writer = null;
		} else {
			Log.i(TAG, "closeLog() called, but no log is open");
		}
		
		for (File f : workingLogPath.listFiles()) {
			if (f.renameTo(new File(uploadLogPath, f.getName()))) {
				Log.i(TAG, "Staged " + f.getName() + " for upload");
			} else {
				Log.e(TAG, "Could not move " + f.getAbsolutePath());
			}
		}
		updateFileStats();
		if (!suppressUpload) uploadThread.trigger(false);
	}
	
	/**
	 * Closes the current log file.
	 * A new log file will be opened automatically when logging is required.
	 */	
	public synchronized void closeLog() {
		closeLog(false);
	}
	
	/**
	 * Shuts down the logging manager.
	 * The object cannot be used after this call.
	 */
	public void close() {
		terminate = true;
		uploadThread.close();
		logIfOpen("TAG", "closing manager");
		closeLog();
		if (logsock != null) logsock.close();
		super.close();
	}
	
	
	
	///////////////////////////////////////////////////////////////////////////////////
	// Upload stuff
	//
	
	public class UploadThread extends Thread {
		private static final String TAG = "Upload";
		private boolean shouldRun = true;
		
		private final AtomicBoolean triggered = new AtomicBoolean();
		/** forced upload (i.e. attempt upload even without WiFi */
		private boolean forced = false;
		@Override
		public void run() {
			while (shouldRun) {
				if (ConsentActivity.hasConsent(service)) {
					setUploading(true);
					File[] files = uploadLogPath.listFiles();
					if (files != null && files.length > 0) {
						for (File f : files) {
							if (hasWifi() || forced) {
								upload(f);
							} else {
								Log.i(TAG, "No WiFi, not uploading");
								break;
							}
						}
					} else {
						Log.i(TAG, "nothing to upload");					
					}
					
					if (hasWifi() || forced) {
						fetchMessages();
					}
					
					if (writer != null && (hasWifi() || forced)) {
						fetchTimeInfo();
					}
					
					setUploading(false);
				} else Log.w(TAG, "Not connecting to server; no consent");
				
				synchronized (triggered) {
					try {
						forced = false;
						// Unless we were triggered during the upload, wait for a trigger
						if (!triggered.get()) {
							setUploadTriggerAlarm();
							triggered.wait();
						}
					} catch (InterruptedException e) {
						// just do the next loop iteration
					}
					triggered.set(false);
				}
			}
		}
		
		private void setUploadTriggerAlarm() {
			if (logUploadAlarmId != -1) alarm.cancelAlarm(logUploadAlarmId);
			logUploadAlarmId = alarm.setAlarm(false, CONFIG.MAX_UPLOAD_PAUSE, new AlarmRunnable() {
				@Override
				public void handleAlarm(int count) {
					trigger(false);
				}
			});
		}
		
		private boolean upload(File file) {
			try {
				Log.i(TAG, "Attempting to upload " + file.getAbsolutePath());
				String hash = CryptoHelper.calculateHMAC(CONFIG.UPLOAD_HMAC_KEY, file);
				
				String requestUri = Uri.parse(CONFIG.UPLOAD_TARGET_URL).buildUpon()
						.appendQueryParameter("node", CONFIG.BT_NAME_PREFIX+service.getNodeID())
						.appendQueryParameter("hash", hash)
						.build().toString();
				
				MultipartEntity multipart = new MultipartEntity();
				multipart.addPart("log", new FileBody(file, file.getName(), "application/octet-stream", "binary"));
								
				// Perform the POST
				String responseText = httpsClient.post(requestUri, multipart);

				// Check result
				if ("OK".equals(responseText)) {
					boolean result = file.delete();
					Log.i(TAG, "success: " + file.getAbsolutePath() + "(deleted: " + result + ")");
					updateFileStats();
					return true;
				} else {
					Log.e(TAG, "upload error: "+ responseText);
					return false;
				}
			} catch (Exception e) {
				Log.e(TAG, "uploadError (Exception) ", e);
				return false;
			} 
		}

		private void fetchMessages() {
			String requestUri = Uri.parse(CONFIG.MESSAGING_URL).buildUpon()
					.appendQueryParameter("node", CONFIG.BT_NAME_PREFIX+service.getNodeID())
					.appendQueryParameter("version", Integer.toString(service.getVersion()))
					.appendQueryParameter("android", Build.VERSION.RELEASE)
					.build().toString();			
			String response = httpsClient.get(requestUri);
			if (response == null) {
				Log.e(TAG, "Failed to fetch messages");
				return;
			}
			String[] parts = response.split("\t");
			if (response.isEmpty() &&
					!(
						parts[0].equals("MSG") ||
						parts[0].equals("KILL") ||
						parts[0].equals("UNKILL")
					)) {
				Log.e(TAG, "Invalid message response" + parts[0]);
				return;
			}

			if (parts[0].equals("KILL")) {
				Log.e("KILLSWITCH", "triggered");
				uploadStateHandler.post(new Runnable() { public void run() { service.setKillswitch(true);} });
			} else if (parts[0].equals("UNKILL")) {
				Log.i("KILLSWITCH", "untriggered");
				uploadStateHandler.post(new Runnable() { public void run() { service.setKillswitch(false);} });
			}

			
			if (parts.length != 3) {
				Log.i(TAG, "No messages");
				setServerMessage(null, null);
				return;
			}

			setServerMessage(parts[1], parts[2]);
		}
		
		

		private void setServerMessage(String url, String text) {
			if (url == null || text == null) {
				if (serverMessageUrl == null && serverMessageText == null) return;
				serverMessageUrl = null;
				serverMessageText = null;
				FileUtils.writeString(service, SERVER_MESSAGE_FILENAME, "");
				uploadStateHandler.sendEmptyMessage(0);	
				return;
			}
			
			// url and text are not null
			if (url.equals(serverMessageUrl) && url.equals(serverMessageText)) return;
			serverMessageUrl = url;
			serverMessageText = text;
			FileUtils.writeString(service, SERVER_MESSAGE_FILENAME, serverMessageUrl+ "\t"+serverMessageText);			
			uploadStateHandler.sendEmptyMessage(0);	
		}

		/**
		 * Triggers immediate uploading if waiting
		 * @param force if true, upload even if no WiFi
		 */
		public void trigger(boolean force) {
			synchronized (triggered) {
				if (force) forced  = true; 
				triggered.set(true);
				triggered.notify();
			}
		}
		
		public void close() {
			shouldRun = false;
			interrupt();
		}
	}
	
	public boolean hasWifi() {
		return connManager.getNetworkInfo(ConnectivityManager.TYPE_WIFI).isConnected();
	}

	public void triggerUpload(boolean force) {
		uploadThread.trigger(force);
	}

	public boolean isUploading() {
		return uploading;
	}

	/**
	 * Sets the uploading state and causes an update.
	 * May be called from any thread.
	 * @param uploading true if upload thread is currently uploading
	 */
	private synchronized void setUploading(boolean uploading) {
		this.uploading = uploading;
		uploadStateHandler.sendEmptyMessage(0);	
	}
	
	/**
	 * Counts files and bytes that have to be uploaded.
	 * May be called from any thread.
	 */
	private void updateFileStats() {
		int count = 0;
		long size = 0;
		long oldest = Long.MAX_VALUE;
		for (File f : uploadLogPath.listFiles()) {
			count++;
			size += f.length();
			if (f.lastModified() < oldest) oldest = f.lastModified(); 
		}
		uploadCount = count;
		uploadSize = size;
		uploadOldest = oldest;
		uploadStateHandler.sendEmptyMessage(0);	
	}

	public int getUploadCount() {
		return uploadCount;
	}

	public long getUploadSize() {
		return uploadSize;
	}

	public long getUploadOldest() {
		return uploadOldest;
	}

	public String getServerMessageUrl() {
		return serverMessageUrl;
	}

	public String getServerMessageText() {
		return serverMessageText;
	}
	
	
	/**
	 * Gets called when the system is shutting down
	 */
	public synchronized void handleShutdownWarning() {
		if (writer == null) return;
		log(TAG, "System shutdown");
		closeLog(true);
	}
	
	@Override
	public void onReceive(Context context, Intent intent) {
		if (intent.getAction().equals(Intent.ACTION_SHUTDOWN)) {
			handleShutdownWarning();
		}
	}

}
