package de.janschejbal.p2pbluetoothtest.bluetooth;

import java.util.ArrayList;
import java.util.UUID;

import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.content.Context;
import android.content.Intent;
import android.os.PowerManager;
import android.os.SystemClock;
import android.os.PowerManager.WakeLock;
import de.janschejbal.p2pbluetoothtest.CONFIG;
import de.janschejbal.p2pbluetoothtest.Logger;
import de.janschejbal.p2pbluetoothtest.protocol.base.ConnectionFactory;
import de.janschejbal.p2pbluetoothtest.util.AlarmHelper;
import de.janschejbal.p2pbluetoothtest.util.AlarmHelper.AlarmRunnable;
import de.janschejbal.p2pbluetoothtest.util.SimpleReceiver;

public class BluetoothCore extends SimpleReceiver {

	private final static String TAG = BluetoothCore.class.getSimpleName();
		
	public static final String BT_SERVICE_NAME = "BluetoothP2PExperiment";
	public static final UUID   BT_SERVICE_UUID = UUID.fromString("f5f20845-3507-413d-9727-abc4a8a15659");
	
	private final BluetoothAdapter bluetooth;
	
	private long lastDiscoveryEndTime;
	
	private final Logger logger;
	
	private final BTCoreStateMachine statemachine;

	private final BluetoothServerThread server;
	private final BluetoothClientThread client;

	private ArrayList<BluetoothDevice> currentDiscoveryList = new ArrayList<BluetoothDevice>();
	private ArrayList<BluetoothDevice> lastDiscoveryList = new ArrayList<BluetoothDevice>();

	public BluetoothCore(Context context, ConnectionFactory connectionFactory, Logger logger) {
		super(context,
					BluetoothAdapter.ACTION_DISCOVERY_STARTED,
					BluetoothAdapter.ACTION_DISCOVERY_FINISHED,
					BluetoothDevice.ACTION_FOUND
				);
		
		this.logger = logger;
		this.bluetooth = BluetoothAdapter.getDefaultAdapter();
		
		statemachine = new BTCoreStateMachine(stateWaitDiscovery);
		
		server = new BluetoothServerThread(connectionFactory, context, logger);
		client = new BluetoothClientThread(connectionFactory, this, logger);
	}

	@Override
	public void onReceive(Context context, Intent intent) {
		
		// A discovery has finished. Write results.
		if (intent.getAction().equals(BluetoothAdapter.ACTION_DISCOVERY_FINISHED)) {
			lastDiscoveryEndTime = SystemClock.elapsedRealtime();
			logger.log(TAG, "Discovery finished");
						
			if (statemachine.getCurrentState() == stateDiscovery ) {
				statemachine.finishState();
			}
					
			return;
		}
		
		// A discovery has started. We reset the timer in case it was external.
		if (intent.getAction().equals(BluetoothAdapter.ACTION_DISCOVERY_STARTED )) {
			if (statemachine.getCurrentState() != stateDiscovery ) {
				logger.log(TAG, "External discovery started");
				statemachine.forceState(stateDiscovery);
			} else {
				logger.log(TAG, "Discovery started");
			}
						
			return;
		}
		
		// We found a device during discovery.
		if (intent.getAction().equals(BluetoothDevice.ACTION_FOUND)) {
			BluetoothDevice dev = intent.getExtras().getParcelable(BluetoothDevice.EXTRA_DEVICE);
			Object signalStrength = intent.getExtras().get(BluetoothDevice.EXTRA_RSSI);
			if (dev.getName() == null) {
				logger.log(TAG, "Device found", "nameless");
			} else if (dev.getName().matches(CONFIG.BT_NAME_PREFIX + "[a-f0-9]{16}")) {
				logger.log(TAG, "Device found", "participant", dev.getName(), signalStrength);
				currentDiscoveryList.add(dev);
			} else if (dev.getName().startsWith(CONFIG.BEACON_PREFIX)) {
				logger.log(TAG, "Device found", "beacon", dev.getName(), signalStrength, dev.getAddress());
			} else {
				logger.log(TAG, "Device found", "non-participant");
			}
			
			return;
		}
	}


	
	public void close() {
		super.close();
		statemachine.close();
		server.close();
		client.close();
	}

	
	/////////////////////////////////////////////////////////////
	// State machine
	//
	
	private State stateDiscovery = new State() {
		@Override
		public String getName() { return "DISCOVERY"; }

		@Override
		public long onEnterState() {
			client.stopCommunication();
			currentDiscoveryList.clear();
			if (!bluetooth.isDiscovering()) bluetooth.startDiscovery();
			return CONFIG.DISCOVERY_TIMEOUT;
		}

		@Override
		public State onFinishState(boolean timeout) {
			if (timeout) bluetooth.cancelDiscovery(); // Note: return value of cancelDiscovery is always false on older Android versions.
			lastDiscoveryList.clear();
			lastDiscoveryList.addAll(currentDiscoveryList);
			return lastDiscoveryList.isEmpty() ? stateWaitDiscovery : stateCommunication;
		}
	};
	
	private State stateCommunication = new State() {
		private WakeLock communicationWakelock = null;
		
		@Override
		public String getName() { return "COMMUNICATION"; }

		@Override
		public long onEnterState() {
			if (communicationWakelock == null) {
				PowerManager pm = (PowerManager) context.getSystemService(Context.POWER_SERVICE);
				communicationWakelock = pm.newWakeLock(PowerManager.PARTIAL_WAKE_LOCK, "P2P_commWakelock");
				communicationWakelock.setReferenceCounted(false);				
			}

			logger.log("WAKELOCK", "comm", "acquire", communicationWakelock.isHeld());
			communicationWakelock.acquire(CONFIG.COMMUNICATION_TIMEOUT+100);
			client.startConnecting(lastDiscoveryList);
			return CONFIG.COMMUNICATION_TIMEOUT;
		}

		@Override
		public State onFinishState(boolean timeout) {
			client.stopConnecting();
			
			// Remaining connections may still exist since they don't get hard-stopped at this point.
			// However, this is the safest place to release the wakelock.
			// We have held the device active for quite a while anyways.
			logger.log("WAKELOCK", "comm", "release", communicationWakelock.isHeld());
			communicationWakelock.release();
			
			return stateWaitDiscovery;
		}
		
	};
	
	/**
	 * Called by the {@link BluetoothClientThread} if it is waiting for devices to connect to.
	 * If we are in the communication state, this indicates all connections have been finished.
	 */
	public void clientConnectionsDone() {
		if (statemachine.getCurrentState() != stateCommunication) return;
		logger.log(TAG, "clientConnectionsDone");
		statemachine.finishState();
	}
	
	private State stateWaitDiscovery = new State() {
		@Override
		public String getName() { return "WAIT"; }

		@Override
		public long onEnterState() {
			long timeSinceLastDiscovery = SystemClock.elapsedRealtime() - lastDiscoveryEndTime;
			return Math.max(0, CONFIG.DISCOVERY_INTERVAL-timeSinceLastDiscovery);
		}

		@Override
		public State onFinishState(boolean timeout) {
			return stateDiscovery;
		}

	};

	private abstract class State {
		public abstract String getName();
		
		/**
		 * Gets called when the state is entered.
		 * @returns Timeout in milliseconds
		 * */
		public abstract long onEnterState();
		
		/**
		 * Gets called when the state finishes. 
		 * @param timeout true if the state timed out/forcefully closed, false otherwise
		 * @returns next state 
		 */
		public abstract State onFinishState(boolean timeout);
	}

	public class BTCoreStateMachine  {
		private final static String TAG = "BTCoreStateMachine";

		private State currentState;
		private long stateCounter = 0;
		
		private final AlarmHelper alarm;
		
		public BTCoreStateMachine(State initialState) {
			alarm = new AlarmHelper(BluetoothCore.this.context);
			enterState(initialState);
		}
		
		public State getCurrentState() {
			return currentState;
		}

		private void timeout(long expectedStateCounter) {
			if (currentState != null && stateCounter == expectedStateCounter) {
				finishState(true);
			} else {
				logger.log(TAG, "ERROR", "unexpected timeout in state", currentState, stateCounter,
						"expected counter", expectedStateCounter);				
			}
		}
		
		/**
		 * Finishes the current state and transitions to the next one.
		 * Check that the current state is the one you want to finish!
		 */
		public void finishState() {
			finishState(false);
		}
		
		private void finishState(boolean timeout) {
			State next = currentState.onFinishState(timeout);
			logger.log(TAG, "State", currentState.getName(), timeout ? "timeout" : "finished");
			enterState(next);			
		}
		
		/**
		 * Set a new state, cancelling (timing out) the current one 
		 * @param newState
		 */
		public void forceState(State newState) {
			logger.log(TAG, "Forcing", newState.getName());
			finishState(true);
			enterState(newState);
		}

		private void enterState(final State next) {
			if (next == null) throw new NullPointerException();
			logger.log(TAG, "State", next.getName(), "entering");
			alarm.cancelAll();
			currentState = next;
			stateCounter++;
			long timeout = next.onEnterState();

			AlarmRunnable newRunnable = new AlarmRunnable() {
				private final long expectedStateCounter = stateCounter;
				
				@Override
				public void handleAlarm(int count) {
					timeout(expectedStateCounter);
				}
			};
			alarm.setAlarm(true, timeout, newRunnable);
		}
		
		public void close() {
			alarm.close();
			if (currentState != null) currentState.onFinishState(true);
			currentState = null;
		}
		
	}

	public boolean serverConnectedTo(String address) {
		return server.isConnectedTo(address);
	}

}
