package de.janschejbal.p2pbluetoothtest.util;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.security.KeyStore;
import java.security.cert.Certificate;
import java.security.cert.CertificateFactory;
import java.util.Collection;

import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.TrustManagerFactory;

import org.apache.http.conn.ssl.AllowAllHostnameVerifier;
import org.apache.http.conn.ssl.StrictHostnameVerifier;
import org.apache.http.entity.mime.MultipartEntity;

import android.content.Context;
import android.util.Log;
import de.janschejbal.p2pbluetoothtest.R;


/**
 * This HTTPS client will trust any certificate specified in res/raw/trusted_certs.txt
 * without any host name verification.
 */
public class HttpsClient {

	private static final String TAG = "HttpsClient";

	private int timeout = 10000;
	private boolean followRedirects = true;
	
	private final SSLSocketFactory sslSocketFactory;
	private final HostnameVerifier hostnameVerifier;

	
	public HttpsClient(Context ctx, boolean checkHostname) {
		
		// Workaround for Android bug causing EOFExceptions on POST
		System.setProperty("http.keepAlive", "false");
		
		hostnameVerifier = checkHostname ?
				new StrictHostnameVerifier()
				: new AllowAllHostnameVerifier();
		
		try {
			// Load certificates
			InputStream is = ctx.getResources().openRawResource(R.raw.trusted_certs);
			CertificateFactory cf = CertificateFactory.getInstance("X509");
			Collection<? extends Certificate> certs = cf.generateCertificates(is);
			FileUtils.closeSilent(is);
	
			// Convert to key store
			KeyStore keyStoreCA = KeyStore.getInstance(KeyStore.getDefaultType());
			keyStoreCA.load(null, null);
			int counter = 0;
			for (Certificate cert : certs) {
				keyStoreCA.setCertificateEntry("trusted"+(counter++), cert);
			}
	
			// Create trust manager
			TrustManagerFactory tmf = TrustManagerFactory.getInstance("X509");
			tmf.init(keyStoreCA);
	
			// Create SSL context
			SSLContext sslcontext = SSLContext.getInstance("TLS");
			sslcontext.init(null, tmf.getTrustManagers(), null);
			
			// Create SSL socket factory...
			sslSocketFactory = sslcontext.getSocketFactory();
		} catch (Exception e) {
			throw new RuntimeException("Failed to initialize " + TAG, e);
		}
	}
	
	/**
	 * Performs a HTTPS GET request.
	 * @param url the HTTPS URL to request
	 * @return the response decoded as an UTF-8 string or null if the request failed (includes non-200 responses)
	 */
	public String get(String url) {
		HttpsURLConnection con = null;
		InputStream in = null;
		try {
			con = getUrlConnection(new URL(url));
			con.setRequestMethod("GET");

			if (con.getResponseCode() != 200) return null;
			in = con.getInputStream();
			return FileUtils.readString(in);
		} catch (IOException e) {
			Log.e(TAG, "GET failed: " + e.toString());
			return null;
		} finally {
			FileUtils.closeSilent(in);
			con.disconnect();
		}		
	}

	/**
	 * Performs a HTTPS POST request.
	 * @param url the HTTPS URL to request
	 * @param multipart the {@link MultipartEntity} to POST to the URL.
	 * @return the response decoded as an UTF-8 string or null if the request failed (includes non-200 responses)
	 */
	public String post(String url, MultipartEntity multipart) {
		HttpsURLConnection con = null;
		OutputStream out = null;
		InputStream in = null;
		try {
			con = getUrlConnection(new URL(url));
			con.setRequestMethod("POST");
			con.setDoOutput(true);
			con.addRequestProperty("Content-type", multipart.getContentType().getValue());
			
			if (multipart.getContentLength() >= 0) {
				con.setFixedLengthStreamingMode((int)multipart.getContentLength());
			} else {
				con.setChunkedStreamingMode(0);
			}

			con.connect();
			out = con.getOutputStream();
			multipart.writeTo(out);
			out.close();
			
			if (con.getResponseCode() != 200) return null;
			in = con.getInputStream();
			return FileUtils.readString(in);
		} catch (IOException e) {
			Log.e(TAG, "POST failed: " + e.toString());
			return null;
		} finally {
			FileUtils.closeSilent(out);
			FileUtils.closeSilent(in);
			con.disconnect();
		}
	}

	private HttpsURLConnection getUrlConnection(URL url) throws IOException {
		if (!url.getProtocol().equals("https")) throw new IOException("URL isn't a https URL");
		HttpsURLConnection con = (HttpsURLConnection)url.openConnection();
		con.setConnectTimeout(timeout);
		con.setReadTimeout(timeout);
		con.setInstanceFollowRedirects(followRedirects);
		con.setSSLSocketFactory(sslSocketFactory);
		con.setHostnameVerifier(hostnameVerifier);
		con.setUseCaches(false);
		return con;
	}
}
