package de.janschejbal.p2pbluetoothtest.broadcast;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.TreeMap;

import android.content.Context;
import de.janschejbal.p2pbluetoothtest.CONFIG;
import de.janschejbal.p2pbluetoothtest.Logger;
import de.janschejbal.p2pbluetoothtest.util.CryptoHelper;
import de.janschejbal.p2pbluetoothtest.util.FileUtils;

/**
 * This class represents the state of all channels (i.e. knows all channels/messages available on this device)
 */
public class BroadcastState {
	/** This map holds the actual state of each channel */
	private final TreeMap<Long, TimestampedChannelState> channelMap;
	
	private final Logger logger;

	private final File savefile;
	private final File savefileTemp;
	
	private final static String TAG = "BroadcastState";

	/**
	 * A random number specifying the instance.
	 * Should only be reset when the application is restarted.
	 * Used to determine validity of elapsedRealtime-based "last changed" timestamps.
	 * 
	 * Each time we meet another node, we tell him our current instance ID and timestamp.
	 * Next time we meet that node, it can request changes since that timestamp.
	 * This can significantly reduce synchronization traffic.
	 * 
	 * The timestamp is only accepted if the instance IDs match.
	 * This prevents confusion when the device is rebooted (and thus elapsedRealtime is reset).
	 */
	private final int instanceId;
	
	{
		// ensure ID is not 0
		int id = CryptoHelper.random.nextInt();
		while (id == 0) id = CryptoHelper.random.nextInt();
		instanceId = id;
	}

	@SuppressWarnings("unchecked")
	protected BroadcastState(Context ctx, Logger logger) {
		this.logger = logger;
		
		this.savefile = new File(ctx.getFilesDir(), "broadcastState.dat");
		this.savefileTemp = new File(ctx.getFilesDir(), "broadcastState.tmp");
		
		ObjectInputStream in = null;
		TreeMap<Long, TimestampedChannelState> tmpChannelMap;
		try {
			if (savefile.length() == 0) throw new Exception("No valid save file");

			in = new ObjectInputStream(new FileInputStream(savefile));
			tmpChannelMap = (TreeMap<Long, TimestampedChannelState>) in.readObject();
		} catch (Exception e) {
			logger.log(TAG, "load failed", e.toString());
			tmpChannelMap = new TreeMap<Long, TimestampedChannelState>();
		}
		channelMap = tmpChannelMap;
		FileUtils.closeSilent(in);
		
		logger.log(TAG, "initialized", tmpChannelMap.size(), getLogString());
	}

	protected synchronized void save() {
		long freshnessThreshold = System.currentTimeMillis() - CONFIG.BROADCAST_CHANNEL_EXPIRY_HARD;
		LinkedList<Long> toRemove = new LinkedList<Long>();
		for (TimestampedChannelState ch : channelMap.values()) {
			if (ch.getLastChangedAbsolute() < freshnessThreshold) {
				toRemove.add(ch.getChannelId());
			}
		}
		
		if (toRemove.size() > 0) {
			for (Long id : toRemove) {
				channelMap.remove(id);
				logger.log(TAG, "purged", CryptoHelper.longToHex(id));
			}
		}
		
		ObjectOutputStream out = null;
		try {
			out = new ObjectOutputStream(new FileOutputStream(savefileTemp));
			out.writeObject(channelMap);
			out.close();
			if (!savefileTemp.renameTo(savefile)) throw new IOException("Move failed");
			logger.log(TAG, "saved", getLogString());
		} catch (Exception e) {
			logger.log(TAG, "save failed");
		}
		FileUtils.closeSilent(out);
	}
	
	/**
	 * Gets a compact version of the current state of all channels as a string
	 * @return a space-separated list of outputs of {@link ChannelState#toString()}
	 */
	public synchronized String getLogString() {
		StringBuilder logString = new StringBuilder();
		for (ChannelState channel : channelMap.values()) {
			logString.append(channel.toString()).append(' ');
		}
		return logString.toString();
	}

	public synchronized ChannelState getChannel(long id) {
		return channelMap.get(id);
	}
	
	private synchronized boolean isBelowChannelLimit() {
		return channelMap.size() < CONFIG.BROADCAST_MAX_CHANNELS;
	}
	
	/**
	 * Checks if a channel already exists in storage
	 * @param channel the channel ID to test
	 * @return true if the channel is known, false if not
	 */
	public synchronized boolean hasChannel(long channel) {
		return channelMap.containsKey(channel);
	}
	
	/**
	 * Checks if a message is in storage (i.e. channel exists and contains message)
	 * @param channelId
	 * @param messageId
	 * @return
	 */
	public synchronized boolean hasMessage(long channelId, int messageId) {
		ChannelState channel = getChannel(channelId);
		if (channel == null) return false;
		return channel.hasMessage(messageId);
	}
	
	/**
	 * Ensures a channel is in storage.
	 * (If it already exists, nothing happens, otherwise an empty bitmap is created.)
	 * @param channel the channel ID (node ID of channel owner)
	 */
	public synchronized void addChannel(long channel) {
		if (hasChannel(channel)) return;
		if (!isBelowChannelLimit()) return; 
		channelMap.put(channel, new TimestampedChannelState(channel));
	}

	/**
	 * Add a message.
	 * @param channel the channel to which the message belongs
	 * @param messageID the message ID of the newly received message
	 */
	public synchronized void addMessage(long channel, int messageID) {
		addChannel(channel);
		channelMap.get(channel).addMessage(messageID);
	}
	
	/**
	 * Returns a list of channels to offer.
	 * If the instanceId is zero or does not match, returns all non-expired channels.
	 * Otherwise, only channels that changed since the timestamp are returned.
	 * The channels are independent clones.
	 * @param instanceId the instance ID on which the timestamp is based, or 0 to get all channels
	 * @param since the timestamp (only channels changed on or after this timestamp are returned)
	 * @return an independent list of independent channel state clones
	 */
	public synchronized ArrayList<ChannelState> getOffer(int instanceId, long since) {
		if (this.instanceId != instanceId || since <= 0) since = Long.MIN_VALUE; 
		ArrayList<ChannelState> channels = new ArrayList<ChannelState>();
		long freshnessThreshold = System.currentTimeMillis() - CONFIG.BROADCAST_CHANNEL_EXPIRY_SOFT; 
		for (TimestampedChannelState channel : channelMap.values()) {
			if (since <= channel.getLastChanged() && freshnessThreshold <= channel.getLastChangedAbsolute() ) {
				channels.add(new ChannelState(channel));
			}
		}
		return channels;
	}
	
	/**
	 * Based on an offer of a remote client, determines which of the offered messages should be requested.
	 * @param offer A list of messages that we are offered
	 * @return an independent list of independent channel states representing the messages to request
	 */
	public synchronized ArrayList<ChannelState> getMessagesToRequest(List<ChannelState> offer) {
		ArrayList<ChannelState> result = new ArrayList<ChannelState>();
		for (ChannelState offeredState : offer) {
			ChannelState localState = getChannel(offeredState.getChannelId());
			if (localState == null && isBelowChannelLimit()) {
				result.add(new ChannelState(offeredState));
			} else {
				ChannelState toRequest = ChannelState.getMissingRight(offeredState, localState);
				if (!toRequest.isEmpty()) result.add(toRequest);
			}	
		}
		return result;
	}

	public int getInstanceId() {
		return instanceId;
	}
	
}
