package de.janschejbal.p2pbluetoothtest.protocol.broadcast;

import java.io.IOException;

import de.janschejbal.p2pbluetoothtest.protocol.base.ProtocolException;
import de.janschejbal.p2pbluetoothtest.protocol.base.ProtocolMessage;
import de.janschejbal.p2pbluetoothtest.protocol.base.ProtocolMessageParser;
import de.janschejbal.p2pbluetoothtest.protocol.base.SafeReader;
import de.janschejbal.p2pbluetoothtest.protocol.base.SafeWriter;

public class EndOfDataMessage implements ProtocolMessage {
	private final int instanceId;
	private final long timestamp;
	
	/**
	 * Creates a new {@link EndOfDataMessage}.
	 * MAY include an instance ID and timestamp (set to 0 otherwise).
	 * If included, the node MAY use these values in the next HelloMessage
	 * and can expect to be provided with all channels that have changed
	 * since the state message of this exchange.
	 * Thus, the timestamp usually represents the creation time when the state message was created.
	 * Note that these values are opaque for the node and can be used in an arbitrary manner.
	 * 
	 * <p>
	 * Parameters must be set to zero or the value in the last HelloMessage
	 * if not all changed states were presented or not all requested messages were sent.
	 * (Exception: Expired messages - since those will never become available again,
	 *  there is no need to make the node re-request them)
	 *  </p>
	 *  
	 * @param instanceId
	 * @param timestamp
	 */
	public EndOfDataMessage(int instanceId, long timestamp) {
		this.instanceId = instanceId;
		this.timestamp = timestamp;
	}

	@Override
	public void send(SafeWriter sink) throws IOException {
		sink.writeHeader(BroadcastProtocolConnection.MESSAGE_END_OF_DATA, 4+8);
		sink.writeInt(instanceId);
		sink.writeLong(timestamp);
	}

	private static final ProtocolMessageParser<EndOfDataMessage> parser = new ProtocolMessageParser<EndOfDataMessage>() {
		@Override
		public EndOfDataMessage parseMessage(int tag, int length, SafeReader source) throws IOException {
			if (tag != BroadcastProtocolConnection.MESSAGE_END_OF_DATA) throw new ProtocolException("Tag mismatch");
			
			int instanceId = source.readInt();
			long timestamp = source.readLong();
			source.chompOverhang();
			return new EndOfDataMessage(instanceId, timestamp);
		}
	};

	public static ProtocolMessageParser<EndOfDataMessage> getParser() {
		return parser;
	}

	public long getTimestamp() {
		return timestamp;
	}

	public int getInstanceId() {
		return instanceId;
	}

}
